/**
 * Notera:
 *
 * Om du vill kunna kompilera koden måste du kopiera mappen "wrap"
 * samt "Makefile" från de givna tentafilerna till samma mapp som du
 * lägger denna filen i.
 *
 * a)
 *
 * Antag att två trådar kör "draw_worker". Den första tråden har
 * "my_line" satt till 3, och den andra har "my_line" satt till
 * 4. Båda trådarna anropar "picture_compute_row" med sin respektive
 * rad och kommer till while-loopen. "line_to_draw" har blivit 3, så
 * den första tråden går igenom while-loopen och ökar
 * "line_to_draw". Detta gör att den andra tråden också kommer igenom
 * while-loopen, och den kan därmed komma till printf-satsen innan den
 * första tråden. I detta fall kommer alltså raderna ut i fel ordning.
 *
 * b)
 *
 * Ja, se markeringar nedan.
 *
 * c)
 *
 * Se markeringar i koden
 *
 * d)
 *
 * Tillagd kod markerad med // (d)
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "picture.h"

/**
 * Följande datastrukturer finns i picture.h:
 * - struct line - ett linjesegment
 * - struct picture - en bild, består av ett antal linjesegment
 *
 * Följande funktioner finns i picture.h:
 *
 * - picture_load: ladda en bild från en fil. Okej att anropa från
 *   flera trådar samtidigt.
 * - picture_free: förstör en bild.
 * - picture_compute_row: beräkna en rad (en viss y-koordinat)
 *   i bilden. Läser endast från picture-objektet.
 */


// Data som skickas till "picture_worker" och delas mellan trådarna
// som ritar upp bilden.
struct draw_data {
    // Bilden som ska ritas upp.
    struct picture *picture;

    // Nästa rad av pixlar som ska genereras.
    int line_to_create;

    // Lås för "line_to_create".
    struct lock create_lock; // (d)

    // Nästa rad av pixlar som ska ritas ut.
    int line_to_draw;

    // Lås för "line_to_draw".
    struct lock draw_lock; // (d)

    // Condition för att vänta på "line_to_draw".
    struct condition draw_cond; // (d)

    // Semafor för att vänta på att alla trådar ska vara klara.
    struct semaphore done; // (d)
};

// Funktion som körs av de trådar som skapas i "draw_picture". Du kan
// anta att denna funktion bara anropas från "draw_picture".
static void draw_worker(void *arg) {
    struct draw_data *data = arg;

    while (true) {
        // Kritisk sektion 1 börjar. Resurs: data->line_to_create
        lock_acquire(&data->create_lock); // (d)

        // Lista ut vilken rad vi ska generera.
        int my_line = data->line_to_create++;

        // Kritisk sektion 1 slutar. Resurs: data->line_to_create
        lock_release(&data->create_lock); // (d)

        if (my_line >= data->picture->height)
            break;

        // Skapa raden.
        char *line = picture_compute_row(data->picture, my_line);

        // Kritisk sektion 2 börjar. Resurs: data->line_to_draw och standard output
        lock_acquire(&data->draw_lock); // (d)

        // Vänta tills det är vår tur att rita upp den.
        // while (data->line_to_draw != my_line)  // <-- busy wait
        //     ;
        while (data->line_to_draw != my_line)              // (d)
            cond_wait(&data->draw_cond, &data->draw_lock); // (d)
        data->line_to_draw++;
        cond_signal(&data->draw_cond, &data->draw_lock);   // (d)

        // Skriv ut den till standard output.
        printf("%s\n", line);

        // Kritisk sektion 2 slutar.
        lock_release(&data->draw_lock);   // (d)

        free(line);
    }

    sema_up(&data->done); // (d)
}

// Rita upp en bild till standard output.
// Notera: du har planer på att utöka denna funktion så att den i
// framtiden också kan spara bilden till en fil. Du vill därmed
// redan nu synkronisera funktionen så att det går att köra den
// parallellt från flera trådar samtidigt.
void draw_picture(struct picture *p, int num_threads) {
    struct draw_data data;
    data.picture = p;
    data.line_to_create = 0;
    data.line_to_draw = 0;

    lock_init(&data.create_lock); // (d)
    lock_init(&data.draw_lock);   // (d)
    cond_init(&data.draw_cond);   // (d)
    sema_init(&data.done, 0);     // (d)

    for (int i = 0; i < num_threads; i++)
        thread_new(&draw_worker, &data);

    // Kritisk sektion 3 börjar. Resurs: data.line_to_draw
    // Notera: denna kritiska sektion försvinner här då vi har använt en semafor i stället.

    // while (data.line_to_draw < p->height)  // <-- busy wait
    //     ;

    // Kritisk sektion 3 slutar.

    for (int i = 0; i < num_threads; i++) // (d)
        sema_down(&data.done);            // (d)
}


/**
 * Huvudprogram.
 *
 * Eventuella ändringar här nedanför kommer att ignoreras vid rättning
 * av uppgiften. Du får däremot ändra nedanför ifall du vill testa din
 * lösning.
 */

int main(int argc, const char **argv) {
    const char *picture = "pintos.txt";
    if (argc > 1)
        picture = argv[1];

    struct picture *p = picture_load(picture);
    if (p == NULL) {
        printf("No picture named %s!\n", picture);
        return 1;
    }

    draw_picture(p, 4);
    picture_free(p);

    return 0;
}
